import asyncio
import logging
import io
import json
import os
import flask
import octoprint.plugin
import octoprint.util
from typing import Any, Dict, List

from octoprint.events import Events

import printnanny_api_client  # beta client
from octoprint.logging.handlers import CleaningTimedRotatingFileHandler

from octoprint_nanny.events import try_handle_event
from octoprint_nanny.manager import WorkerManager
from octoprint_nanny.utils.printnanny_os import (
    issue_txt,
    printnanny_device,
    printnanny_config,
    printnanny_user,
    printnanny_device,
    janus_edge_hostname,
    janus_edge_api_token,
    etc_os_release,
)

logger = logging.getLogger("octoprint.plugins.octoprint_nanny")


def configure_logger(logger, logfile_path):

    file_logging_handler = CleaningTimedRotatingFileHandler(
        logfile_path,
        when="D",
        backupCount=7,
    )
    file_logging_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s - %(name)s - %(module)s - %(thread)d - %(levelname)s - %(message)s"
        )
    )
    file_logging_handler.setLevel(logging.DEBUG)

    logger.addHandler(file_logging_handler)

    logger.info(f"Logger file handler added {file_logging_handler}")


DEFAULT_API_URL = os.environ.get(
    "OCTOPRINT_NANNY_API_URL", "https://print-nanny.com/api/"
)
DEFAULT_WS_URL = os.environ.get("OCTOPRINT_NANNY_WS_URL", "wss://print-nanny.com/ws/")
DEFAULT_SNAPSHOT_URL = os.environ.get(
    "OCTOPRINT_NANNY_SNAPSHOT_URL", "http://localhost:8080/?action=snapshot"
)

DEFAULT_MQTT_BRIDGE_PORT = os.environ.get("OCTOPRINT_NANNY_MQTT_BRIDGE_PORT", 443)
DEFAULT_MQTT_BRIDGE_HOSTNAME = os.environ.get(
    "OCTOPRINT_NANNY_MQTT_HOSTNAME", "mqtt.2030.ltsapis.goog"
)
DEFAULT_MQTT_ROOT_CERTIFICATE_URL = "https://pki.google.com/roots.pem"
BACKUP_MQTT_ROOT_CERTIFICATE_URL = "https://pki.goog/gsr4/GSR4.crt"

Events.PRINT_PROGRESS = "PrintProgress"


class OctoPrintNannyPlugin(
    octoprint.plugin.SettingsPlugin,
    octoprint.plugin.AssetPlugin,
    octoprint.plugin.TemplatePlugin,
    octoprint.plugin.WizardPlugin,
    octoprint.plugin.BlueprintPlugin,
    octoprint.plugin.StartupPlugin,
    octoprint.plugin.EventHandlerPlugin,
    octoprint.plugin.EnvironmentDetectionPlugin,
    octoprint.plugin.ProgressPlugin,
    octoprint.plugin.ShutdownPlugin,
    octoprint.plugin.RestartNeedingPlugin,
):

    octoprint_event_prefix = "plugin_octoprint_nanny_"
    plugin_identifier = "octoprint_nanny"
    VERBOSE_EVENTS = [Events.Z_CHANGE]

    def __init__(self, *args, **kwargs):
        self._log_path = None
        self.worker_manager = WorkerManager(plugin=self)
        super().__init__(*args, **kwargs)

    def _test_api_auth(self, auth_token: str, api_url: str):
        response = asyncio.run_coroutine_threadsafe(
            self._test_api_auth_async(auth_token, api_url), self.worker_manager.loop
        )
        if response.exception():
            return response.exception()
        else:
            return response.result()

    async def sync_printer_profiles(self, **kwargs) -> bool:
        octoprint_device_id = self.get_setting("octoprint_device_id")
        if octoprint_device_id is None:
            return False
        logger.info(
            f"Syncing printer profiles for octoprint_device_id={octoprint_device_id}"
        )
        printer_profiles = self._printer_profile_manager.get_all()

        # on sync, cache a local map of octoprint id <-> print nanny id mappings for debugging
        id_map: Dict[str, Dict[str, int]] = {"octoprint": {}, "octoprint_nanny": {}}
        for profile_id, profile in printer_profiles.items():
            try:
                created_profile = await self.worker_manager.plugin.settings.rest_client.update_or_create_printer_profile(
                    profile, octoprint_device_id
                )
                id_map["octoprint"][profile_id] = created_profile.id
                id_map["octoprint_nanny"][created_profile.id] = profile_id
            except printnanny_api_client.exceptions.ApiException as e:
                # octoprint device was deleted remotely
                if e.status == 400:
                    try:
                        res = json.loads(e.body)
                        if res.get("octoprint_device") is not None:
                            self._reset_octoprint_device()
                    except ValueError as e2:  # not all responses are JSON serializable right now (e.g. django default responses)
                        pass
                logger.error(f"Error syncing printer profiles {e.body}")
                return False
            except asyncio.TimeoutError as e:
                logger.error(f"Connection to Print Nanny REST API timed out")
                return False

        logger.info(f"Synced {len(printer_profiles)} printer_profile")

        filename = os.path.join(
            self.get_plugin_data_folder(), "printer_profile_id_map.json"
        )
        with io.open(filename, "w+", encoding="utf-8") as f:
            json.dump(id_map, f)
        logger.info(
            f"Wrote id map for {len(printer_profiles)} printer profiles to {filename}"
        )
        return True

    ##
    ## Octoprint api routes + handlers
    ##
    @octoprint.plugin.BlueprintPlugin.route("/createBackup", methods=["POST"])
    def create_backup(self):
        helpers = self._plugin_manager.get_helpers("backup", "create_backup")

        if helpers and "create_backup" in helpers:
            backup_file = helpers["create_backup"](exclude=["timelapse"])
            logger.info("Created backup file")
            return flask.json.jsonify({"ok": 1})
        else:
            logger.error("Plugin manager failed to get backup helper")
            raise Exception("Plugin manager failed to get backup helper")

    def register_custom_events(self) -> List[str]:
        remote_commands = [
            "remote_command_received",
            "remote_command_failed",
            "remote_command_success",
            "connect_test_mqtt_pong_success",
            "nnstreamer_start",
            "nnstreamer_stop",
        ]
        return remote_commands

    def on_shutdown(self):
        logger.info("Running on_shutdown handler")

    def on_startup(self, *args, **kwargs):
        logger.info("Running on_startup handler args=%s kwargs=%s", args, kwargs)

    def on_after_startup(self, *args, **kwargs):
        logger.info("Running on_after_startup handler args=%s kwargs=%s", args, kwargs)
        configure_logger(logger, self._settings.get_plugin_logfile_path())

    def on_event(self, event: str, payload: Dict[Any, Any]):
        events_enabled = self._settings.get(["events_enabled"])
        config = self._settings.get(["printnanny_config"])
        if config is None:
            logger.warning(
                "PrintNanny OS not detected or device is not registered. Ignoring event %s",
                event,
            )
            return
        try_handle_event(event, payload, config=config, events_enabled=events_enabled)

    def on_environment_detected(self, environment, *args, **kwargs):
        logger.info(
            "Running on_environment_detectedp handler args=%s kwargs=%s", args, kwargs
        )

        self._octoprint_environment = environment

    def on_settings_initialized(self):
        """
        Called after plugin initialization
        """
        logger.info("Running on_settings_initialized handler")

        self._log_path = self._settings.get_plugin_logfile_path()

    ## Progress plugin

    def on_print_progress(self, storage, path, progress):
        octoprint_job = self._printer.get_current_job()
        pass

    ## SettingsPlugin mixin
    def get_settings_defaults(self):
        DEFAULT_SETTINGS = dict(
            backup_auto=False,
            analytics_enabled=False,
            events_enabled=False,
            wizard_complete=-1,
        )
        return DEFAULT_SETTINGS

    ## Template plugin

    def get_template_vars(self):
        custom = {
            "urls": {
                "getting_started_guide": "https://bitsy-ai.notion.site/Getting-Started-with-Print-Nanny-OS-817bc65297ff44a085120c663dced5f3",
                "discord_invite": "https://discord.gg/sf23bk2hPr",
                "cloud": "https://printnanny.ai",
            },
            "printnanny_user": printnanny_user(),
            "printnanny_device": printnanny_device(),
            "janus_edge_hostname": janus_edge_hostname(),
            "janus_edge_api_token": janus_edge_api_token(),
            "issue_txt": issue_txt(),
            "etc_os_release": etc_os_release(),
            "config": printnanny_config(),
        }
        return custom

    ## Wizard plugin mixin

    def get_wizard_version(self):
        return 0

    def is_wizard_required(self):
        return self._settings.get(["wizard_complete"]) < self.get_wizard_version()

    ##~~ AssetPlugin mixin

    def get_assets(self):
        # Define your plugin's asset files to automatically include in the
        # core UI here.
        return dict(
            js=["js/nanny.js", "vendor/janus/janus.js", "vendor/janus/settings.js"],
            css=["css/printnanny.css"],
            less=["less/nanny.less"],
            img=["img/wizard_example.jpg"],
        )

    ##~~ Softwareupdate hook
    def get_update_information(self):
        # Define the configuration for your plugin to use with the Software Update
        # Plugin here. See https://docs.octoprint.org/en/master/bundledplugins/softwareupdate.html
        # for details.
        return dict(
            octoprint_nanny=dict(
                displayName=self._plugin_name,
                displayVersion=self._plugin_version,
                # version check: github repository
                type="github_release",
                user="bitsy-ai",
                repo="octoprint-nanny-plugin",
                current=self._plugin_version,
                # update method: pip
                pip="https://github.com/bitsy-ai/octoprint-nanny-plugin/archive/{target_version}.zip",
                stable_branch=dict(
                    name="Stable Channel", branch="main", commitish=["main"]
                ),
                prerelease_branches=[
                    dict(
                        name="Prerelease Channel",
                        branch="devel",
                        commitish=["main", "devel"],
                    )
                ],
            ),
        )



# type: ignore
from __future__ import absolute_import

__plugin_name__ = "OctoPrint-Nanny"

__plugin_pythoncompat__ = ">=3,<4"  # only python 3

__plugin_version__ = "0.11.2"


def __plugin_load__():
    from .plugins import OctoPrintNannyPlugin

    global __plugin_implementation__
    __plugin_implementation__ = OctoPrintNannyPlugin()
    __plugin_implementation__.version = __plugin_version__
    global __plugin_hooks__
    __plugin_hooks__ = {
        "octoprint.plugin.softwareupdate.check_config": __plugin_implementation__.get_update_information,
        # "octoprint.timelapse.capture.post": __plugin_implementation__.on_timelapse_capture,
        "octoprint.events.register_custom_events": __plugin_implementation__.register_custom_events,
        # "octoprint.server.http.routes": __plugin_implementation__.register_custom_routes
    }



import logging
import socket
import json
from typing import Dict, Any, Optional

# TODO descriminator models not getting generated in python?
import printnanny_api_client.models
from octoprint_nanny.exceptions import SetupIncompleteError
from octoprint_nanny.utils.printnanny_os import printnanny_config

logger = logging.getLogger(__name__)
# see available events: https://docs.octoprint.org/en/master/events/index.html#id5

PUBLISH_EVENTS = [
    "Startup",  # server
    "Shutdown",  # server
    "PrintProgress",  # printer comms
    "Connecting",  # printer comms
    "Connected",  # printer comms
    "Disconnecting",  # printer comms
    "Disconnected",  # printer comms
    "Error",  # printer comms
    "PrintStarted",  # print job
    "PrintFailed",  # print job
    "PrintDone",  # print job
    "PrintCancelling",  # print job
    "PrintCancelled",  # print job
    "PrintPaused",  # print job
    "PrintResumed",  # print job
]


def should_publish_event(event: str) -> bool:
    return event in PUBLISH_EVENTS


def event_request(
    event: str, payload: Dict[Any, Any], device: int, octoprint_install: int
) -> printnanny_api_client.models.OctoPrintEventRequest:
    return printnanny_api_client.models.OctoPrintEventRequest(
        model="OctoPrintEvent",
        source=printnanny_api_client.models.EventSource.OCTOPRINT,
        event_name=event,
        payload=payload,
        device=device,
        octoprint_install=octoprint_install,
    )


def try_write_socket(
    request: printnanny_api_client.models.OctoPrintEventRequest, events_socket: str
) -> None:
    data = json.dumps(request.to_dict()).encode("utf-8")
    logger.debug(
        "Publishing data %s to socket %s",
        data,
        events_socket,
    )
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as client:
        client.connect(events_socket)
        client.sendall(data)
        client.close()
        logger.debug("Closed socket %s", events_socket)
    return


def try_publish_event(
    event: str, payload: Dict[Any, Any], config: Dict[Any, Any]
) -> Optional[printnanny_api_client.models.OctoPrintEventRequest]:
    if should_publish_event(event):
        device = config.get("device", {}).get("id")
        octoprint_install = config.get("octoprint_install", {}).get("id")
        socket = config.get("events_socket")
        if device is None:
            raise SetupIncompleteError("printnanny_config.device is not set")
        if octoprint_install is None:
            raise SetupIncompleteError("printnanny_config.octoprint_install is not set")
        if socket is None:
            raise SetupIncompleteError("printnanny_config.events_socket is not set")
        try:
            req = event_request(event, payload, device, octoprint_install)
            try_write_socket(req, config["events_socket"])
            return req
        except Exception as e:
            logger.error("Error publishing event=%s error=%s", event, e)
            return None
    else:
        logger.debug("Ignoring event %s", event)
        return None


def try_handle_event(
    event: str,
    payload: Dict[Any, Any],
    config: Dict[Any, Any],
    events_enabled: bool = True,
) -> Optional[printnanny_api_client.models.OctoPrintEventRequest]:
    if events_enabled:
        return try_publish_event(event, payload, config)
    logger.debug(
        "Skipping publish for event=%s, events_enabled=%s", event, events_enabled
    )
    return None



"""
This module bundles commonly used utility methods or helper classes that are used in multiple places within
OctoPrint's source code.

Vendored in OctoPrint-Nanny to avoid native build dependency on OctoPrint (and OctoPrint's large dependency tree) in Yocto build system
"""


__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"

import copy
import pickle


def fast_deepcopy(obj):
    # the best way to implement this would be as a C module, that way we'd be able to use
    # the fast path every time.
    try:
        # implemented in C and much faster than deepcopy:
        # https://stackoverflow.com/a/29385667
        return pickle.loads(pickle.dumps(obj, pickle.HIGHEST_PROTOCOL))
    except (AttributeError, pickle.PicklingError):
        # fall back when something unpickable is found
        return copy.deepcopy(obj)


def dict_merge(a, b, leaf_merger=None, in_place=False):
    """
    Recursively deep-merges two dictionaries.
    Based on https://www.xormedia.com/recursively-merge-dictionaries-in-python/
    Example::
        >>> a = dict(foo="foo", bar="bar", fnord=dict(a=1))
        >>> b = dict(foo="other foo", fnord=dict(b=2, l=["some", "list"]))
        >>> expected = dict(foo="other foo", bar="bar", fnord=dict(a=1, b=2, l=["some", "list"]))
        >>> dict_merge(a, b) == expected
        True
        >>> dict_merge(a, None) == a
        True
        >>> dict_merge(None, b) == b
        True
        >>> dict_merge(None, None) == dict()
        True
        >>> def leaf_merger(a, b):
        ...     if isinstance(a, list) and isinstance(b, list):
        ...         return a + b
        ...     raise ValueError()
        >>> result = dict_merge(dict(l1=[3, 4], l2=[1], a="a"), dict(l1=[1, 2], l2="foo", b="b"), leaf_merger=leaf_merger)
        >>> result.get("l1") == [3, 4, 1, 2]
        True
        >>> result.get("l2") == "foo"
        True
        >>> result.get("a") == "a"
        True
        >>> result.get("b") == "b"
        True
        >>> c = dict(foo="foo")
        >>> dict_merge(c, {"bar": "bar"}) is c
        False
        >>> dict_merge(c, {"bar": "bar"}, in_place=True) is c
        True
    Arguments:
        a (dict): The dictionary to merge ``b`` into
        b (dict): The dictionary to merge into ``a``
        leaf_merger (callable): An optional callable to use to merge leaves (non-dict values)
        in_place (boolean): If set to True, a will be merged with b in place, meaning a will be modified
    Returns:
        dict: ``b`` deep-merged into ``a``
    """

    if a is None:
        a = {}
    if b is None:
        b = {}

    if not isinstance(b, dict):
        return b

    if in_place:
        result = a
    else:
        result = fast_deepcopy(a)

    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(
                result[k], v, leaf_merger=leaf_merger, in_place=in_place
            )
        else:
            merged = None
            if k in result and callable(leaf_merger):
                try:
                    merged = leaf_merger(result[k], v)
                except ValueError:
                    # can't be merged by leaf merger
                    pass

            if merged is None:
                merged = fast_deepcopy(v)

            result[k] = merged
    return result



import socket


class SetupIncompleteError(Exception):
    def __init__(
        self,
        msg=f"PrintNanny registration is incomplete. Please visit http://{socket.gethostname()}/ to finish setup.",
        *args,
        **kwargs,
    ):
        super().__init__(msg, *args, **kwargs)



import asyncio
import concurrent
import logging
import queue
import threading

from octoprint.events import Events

from octoprint_nanny.clients.rest import API_CLIENT_EXCEPTIONS

Events.PRINT_PROGRESS = "PrintProgress"
logger = logging.getLogger("octoprint.plugins.octoprint_nanny.manager")


class WorkerManager:
    """
    Manages an asynio event loop running in a separate ThreadPool
    """

    def __init__(self, plugin):

        self.event_loop_thread = threading.Thread(target=self._event_loop_worker)
        self.event_loop_thread.daemon = True
        self.event_loop_thread.start()
        # outbound telemetry messages to MQTT bridge
        self.mqtt_send_queue: queue.Queue = queue.Queue()
        # inbound MQTT command and config messages from MQTT bridge
        self.mqtt_receive_queue: queue.Queue = queue.Queue()

        self.plugin = plugin

    def _event_loop_worker(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_default_executor(concurrent.futures.ThreadPoolExecutor(max_workers=4))
        self.loop = loop
        return loop.run_forever()

    async def on_user_logged_in(self, **kwargs):
        try:
            await self.plugin.sync_printer_profiles()
            await self.plugin.sync_device_metadata()
        except API_CLIENT_EXCEPTIONS:  # rest_client.py contains backoff and giveup logging hanlers
            pass



from os import environ
from typing import Optional, Any, Dict
import logging
import json
import subprocess
import socket

logger = logging.getLogger("octoprint.plugins.octoprint_nanny.utils")

PRINTNANNY_BIN = environ.get("PRINTNANNY_BIN", "/usr/bin/printnanny")
PRINTNANNY_PROFILE = environ.get("PRINTNANNY_PROFILE", "default")


def printnanny_user() -> Optional[Dict[Any, Any]]:
    cmd = [PRINTNANNY_BIN, "config", "get", "user", "-F", "json"]
    try:
        p = subprocess.run(cmd, capture_output=True)
    except FileNotFoundError as e:
        logger.error(e)
        return None
    stdout = p.stdout.decode("utf-8")
    stderr = p.stderr.decode("utf-8")
    if p.returncode != 0:
        logger.warning(
            f"Failed to get printnanny user cmd={cmd} returncode={p.returncode} stdout={stdout} stderr={stderr}"
        )
        return None
    logger.info(f"Logged in as printnanny user={stdout}")
    try:
        user = json.loads(stdout)
        return user
    except json.JSONDecodeError as e:
        logger.error(e)
        logger.error(f"Failed to decode printnanny config: {stdout}")
        return None


def printnanny_device() -> Optional[Dict[Any, Any]]:
    cmd = [PRINTNANNY_BIN, "config", "get", "device", "-F", "json"]
    try:
        p = subprocess.run(cmd, capture_output=True)
    except FileNotFoundError as e:
        logger.error(e)
        return None
    stdout = p.stdout.decode("utf-8")
    stderr = p.stderr.decode("utf-8")
    if p.returncode != 0:
        logger.warning(
            f"Failed to get printnanny device cmd={cmd} returncode={p.returncode} stdout={stdout} stderr={stderr}"
        )
        return None
    logger.info(f"Authenticated with device={stdout}")
    try:
        user = json.loads(stdout)
        return user
    except json.JSONDecodeError as e:
        logger.error(e)
        logger.error(f"Failed to decode printnanny device={stdout}")
        return None


def printnanny_config() -> Optional[Dict[str, Any]]:
    cmd = [PRINTNANNY_BIN, "config", "show", "-F", "json"]
    try:
        p = subprocess.run(cmd, capture_output=True)
    except FileNotFoundError as e:
        logger.error(e)
        return None
    stdout = p.stdout.decode("utf-8")
    stderr = p.stderr.decode("utf-8")
    if p.returncode != 0:
        logger.warning(
            f"Failed to get printnanny config cmd={cmd} returncode={p.returncode} stdout={stdout} stderr={stderr}"
        )
        return None
    try:
        config = json.loads(stdout)
        logger.info(f"Read printnanny config: {config}")
        return config
    except json.JSONDecodeError as e:
        logger.error(e)
        logger.error(f"Failed to decode printnanny config: {stdout}")
        return None


def janus_edge_hostname() -> str:
    return environ.get("JANUS_EDGE_HOSTNAME", "localhost")


def janus_edge_api_token() -> str:
    return environ.get("JANUS_EDGE_API_TOKEN", "janustoken")


def issue_txt() -> str:
    """
    Captured the contents of /boot/issue.txt as plain text
    """
    try:
        result = open("/boot/issue.txt", "r").read().strip()
    except Exception as e:
        logger.error("Failed to read /boot/issue.txt %s", e)
        result = "Failed to read /boot/issue.txt"
    return result


def etc_os_release() -> Dict[str, str]:
    """
    Captures the contents of /etc/os-release as a dictionary
    """
    f = open("/etc/os-release", "r").read()
    result = dict(ID="unknown")
    try:
        lines = f.strip().split("\n")
        for line in lines:
            k, v = line.split("=")
            result[k] = v
    except Exception as e:
        logger.error("Error parsing contents of /etc/os-release %s", e)
    return result



import aiohttp
import logging
import urllib.parse
import backoff
import os

import printnanny_api_client
from printnanny_api_client import ApiClient as AsyncApiClient

from printnanny_api_client.api.users_api import UsersApi


logger = logging.getLogger("octoprint.plugins.octoprint_nanny.clients.rest")

API_CLIENT_EXCEPTIONS = (
    printnanny_api_client.exceptions.ApiException,
    aiohttp.client_exceptions.ClientError,
)
MAX_BACKOFF_TIME = int(os.environ.get("OCTOPRINT_NANNY_MAX_BACKOFF_TIME", 120))

logger.info(f"OCTOPRINT_NANNY_MAX_BACKOFF_TIME={MAX_BACKOFF_TIME}")


def fatal_code(e) -> bool:
    """
    Returns True if error code is fatal and should not be retried
    """
    if isinstance(e, aiohttp.ClientConnectionError):
        return False
    return 400 <= e.response.status_code < 500


def backoff_hdlr(details):
    logger.warning(
        "Backing off {wait:0.1f} seconds afters {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )


def giveup_hdlr(details):
    logger.error(
        "Giving up {elapsed:0.1f} seconds afters {tries} tries "
        "calling function {target} with args {args} and kwargs "
        "{kwargs}".format(**details)
    )


class RestAPIClient:
    """
    webapp rest API calls and retry behavior
    """

    def __init__(self, auth_token: str, api_url: str):
        self.api_url = api_url
        self.auth_token = auth_token

    @property
    def _api_config(self):
        parsed_uri = urllib.parse.urlparse(self.api_url)
        host = f"{parsed_uri.scheme}://{parsed_uri.netloc}"
        config = printnanny_api_client.Configuration(host=host)

        config.access_token = self.auth_token
        return config

    @backoff.on_exception(
        backoff.expo,
        aiohttp.ClientConnectionError,
        logger=logger,
        max_time=MAX_BACKOFF_TIME,
        jitter=backoff.random_jitter,
        giveup=fatal_code,
        on_backoff=backoff_hdlr,
        on_giveup=giveup_hdlr,
    )
    async def get_user(self):
        async with AsyncApiClient(self._api_config) as api_client:
            api_instance = UsersApi(api_client=api_client)
            user = await api_instance.users_me_retrieve()
            return user

    # @backoff.on_exception(
    #     backoff.expo,
    #     aiohttp.ClientConnectionError,
    #     logger=logger,
    #     max_time=MAX_BACKOFF_TIME,
    #     jitter=backoff.random_jitter,
    #     giveup=fatal_code,
    #     on_backoff=backoff_hdlr,
    #     on_giveup=giveup_hdlr,
    # )
    # async def update_or_create_printer_profile(
    #     self, printer_profile, octoprint_device_id
    # ):
    #     """
    #     https://github.com/OctoPrint/OctoPrint/blob/f67c15a9a47794a68be9aed4f2d5a12a87e70179/src/octoprint/printer/profile.py#L46
    #     """

    #     async with AsyncApiClient(self._api_config) as api_client:
    #         # printer profile
    #         # TODO re-enable with OctoPrintApi module
    #         # api_instance = RemoteControlApi(api_client=api_client)

    #         # cooerce duck-typed fields
    #         if type(printer_profile["volume"]["custom_box"]) is bool:
    #             volume_custom_box = {}
    #         else:
    #             volume_custom_box = printer_profile["volume"]["custom_box"]

    #         request = PrinterProfileRequest(
    #             octoprint_device=octoprint_device_id,
    #             octoprint_key=printer_profile["id"],
    #             axes_e_inverted=printer_profile["axes"]["e"]["inverted"],
    #             axes_x_inverted=printer_profile["axes"]["x"]["inverted"],
    #             axes_y_inverted=printer_profile["axes"]["y"]["inverted"],
    #             axes_z_inverted=printer_profile["axes"]["z"]["inverted"],
    #             axes_e_speed=printer_profile["axes"]["e"]["speed"],
    #             axes_x_speed=printer_profile["axes"]["x"]["speed"],
    #             axes_y_speed=printer_profile["axes"]["y"]["speed"],
    #             axes_z_speed=printer_profile["axes"]["z"]["speed"],
    #             extruder_count=printer_profile["extruder"]["count"],
    #             extruder_nozzle_diameter=printer_profile["extruder"]["nozzleDiameter"],
    #             extruder_shared_nozzle=printer_profile["extruder"]["sharedNozzle"],
    #             name=printer_profile["name"],
    #             model=printer_profile["model"],
    #             heated_bed=printer_profile["heatedBed"],
    #             heated_chamber=printer_profile["heatedChamber"],
    #             volume_custom_box=volume_custom_box,
    #             volume_depth=printer_profile["volume"]["depth"],
    #             volume_formfactor=printer_profile["volume"]["formFactor"],
    #             volume_height=printer_profile["volume"]["height"],
    #             volume_origin=printer_profile["volume"]["origin"],
    #             volume_width=printer_profile["volume"]["width"],
    #         )
    #         printer_profile = await api_instance.printer_profiles_update_or_create(
    #             request
    #         )
    #         return printer_profile

    async def create_backup(
        self, hostname: str, name: str, octoprint_version: str, file: str
    ):
        async with AsyncApiClient(self._api_config) as api_client:
            api_instance = printnanny_api_client.OctoprintApi(api_client=api_client)

            backup = await api_instance.octoprint_backups_create(
                hostname,
                name,
                octoprint_version,
                file,
            )
            return backup



import json
import base64
from enum import Enum
import datetime
from io import BytesIO
import base64


class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime.date, datetime.datetime)):
            return obj.isoformat()
        elif isinstance(obj, BytesIO):
            obj.seek(0)
            obj = obj.read()
            return base64.b64encode(obj).decode()
        elif isinstance(obj, bytes):
            return obj.decode()
        elif isinstance(obj, Enum):
            return obj.value
        return json.JSONEncoder.default(self, obj)
